from cpython.ref cimport PyObject
from gimpact cimport *
import numpy as np
cimport numpy as np

cdef object __create_uninitialized__ = object()


cdef class AABB:
    """Represents an axis-aligned bounding boxes (AABB). If the AABB is returned by
    an AABBSet, the AABB still "belongs to" to that AABBSet, so modifying it will 
    alter the AABBSet; make a copy of it if you wish to modify it seperately.
    
    :param min_x: minimum of x axis
    :type min_x: float
    :param max_x: maximum of x axis
    :type max_x: float
    :param min_y: minimum of y axis
    :type min_y: float
    :param max_y: maximum of y axis
    :type max_y: float
    :param min_z: minimum of z axis
    :type min_z: float
    :param max_z: maximum of z axis
    :type max_z: float
    """
    cdef object __weakref__
    cdef aabb3f _aabb
    cdef aabb3f* _src_aabb
    cdef GIM_AABB_SET* _aabb_set
    
    def __cinit__(self, min_x, max_x, min_y, max_y, min_z, max_z):
        self._aabb.minX = min_x
        self._aabb.maxX = max_x
        self._aabb.minY = min_y
        self._aabb.maxY = max_y
        self._aabb.minZ = min_z
        self._aabb.maxZ = max_z
        self._aabb_set = NULL
        self._src_aabb = NULL
    
    def __getitem__(self, index):
        return self.bounds[index]

    def __setitem__(self, index, value):
        bounds = list(self.bounds)
        bounds[index] = value
        self.bounds = bounds

    cdef updateLocalAABB(self):
        if self._aabb_set is not NULL and self._aabb_set.m_count != 0 and self._src_aabb is not NULL:
            AABB_COPY(self._aabb, self._src_aabb[0])

    @property
    def bounds(self):
        """Gets and Sets Bounds of the AABB as a tuple (min_x, max_x, min_y, max_y, min_z, max_z)"""
        self.updateLocalAABB()            
        return (self._aabb.minX, self._aabb.maxX, self._aabb.minY, self._aabb.maxY, self._aabb.minZ, self._aabb.maxZ)

    @bounds.setter
    def bounds(self, value):
        self._aabb.minX, self._aabb.maxX, self._aabb.minY, self._aabb.maxY, self._aabb.minZ, self._aabb.maxZ = value
        self.updateAABBSet()

    cdef updateAABBSet(self):
        if self._aabb_set is not NULL and self._aabb_set.m_count != 0 and self._src_aabb is not NULL:
            AABB_COPY(self._src_aabb[0], self._aabb)
            gim_aabbset_update(self._aabb_set)
        
    @property
    def min_x(self):
        """Gets and Sets minimum x-axis value of the AABB"""
        return self.__getitem__(0)

    @min_x.setter
    def min_x(self, value):
        self.__setitem__(0, value)
        
    @property
    def max_x(self):
        """Gets and Sets maximum x-axis value of the AABB"""
        return self.__getitem__(1)

    @max_x.setter
    def max_x(self, value):
        self.__setitem__(1, value)
        
    @property
    def min_y(self):
        """Gets and Sets minimum x-axis value of the AABB"""
        return self.__getitem__(2)

    @min_y.setter
    def min_y(self, value):
        self.__setitem__(2, value)
        
    @property
    def max_y(self):
        """Gets and Sets maximum y-axis value of the AABB"""
        return self.__getitem__(3)

    @max_y.setter
    def max_y(self, value):
        self.__setitem__(3, value)
        
    @property
    def min_z(self):
        """Gets and Sets minimum z-axis value of the AABB"""
        return self.__getitem__(4)

    @min_z.setter
    def min_z(self, value):
        self.__setitem__(4, value)  
        
    @property
    def max_z(self):
        """Gets and Sets maximum z-axis value of the AABB"""
        return self.__getitem__(5)

    @max_z.setter
    def max_z(self, value):
        self.__setitem__(5, value)  

    def intersects(self, aabb):
        """Checks if the AABB intersects with a given AABB.

        :param aabb: bounding box
        :type aabb: AABB
        :return: flag indicating if the boxes intersect
        :rtype: bool
        """
        cdef char intersected = 0
        cdef aabb3f aabb2
        aabb2.minX, aabb2.maxX, aabb2.minY, aabb2.maxY, aabb2.minZ, aabb2.maxZ = aabb.bounds
        self.updateLocalAABB()
        AABBCOLLISION(intersected, self._aabb, aabb2) 
        return True if intersected == 1 else False

    def intersection(self, aabb):
        """Finds intersection between the AABB and a given AABB. The resulting 
        intersection will an invalid bounding box (i.e minimum will be greater 
        than maximum in the non-overlapping axis) if both boxes do not intersect.

        :param aabb: bounding box
        :type aabb: AABB
        :return: intersection of boxes
        :rtype: AABB
        """
        result = AABB(0, 0, 0, 0, 0, 0)
        cdef aabb3f aabb2
        aabb2.minX, aabb2.maxX, aabb2.minY, aabb2.maxY, aabb2.minZ, aabb2.maxZ = aabb.bounds
        self.updateLocalAABB()
        BOXINTERSECTION(self._aabb, aabb2, result._aabb)
        return result

    def merge(self, aabb):
        """Merge the AABB with a given AABB.

        :param aabb: bounding box
        :type aabb: AABB
        """ 
        result = AABB(0, 0, 0, 0, 0, 0)
        cdef aabb3f aabb2
        aabb2.minX, aabb2.maxX, aabb2.minY, aabb2.maxY, aabb2.minZ, aabb2.maxZ = aabb.bounds
        self.updateLocalAABB() 
        MERGEBOXES(self._aabb, aabb2)
        self.updateAABBSet()

    def clone(self):
        """clones the AABB.
        
        :return: AABB clone
        :rtype: AABB
        """
        self.updateLocalAABB() 
        return AABB(self._aabb.minX, self._aabb.maxX, self._aabb.minY, self._aabb.maxY, self._aabb.minZ, self._aabb.maxZ)

    def _id(self):
        return PyLong_FromVoidPtr(<void*>self._aabb_set)

    def __str__(self):
        self.updateLocalAABB() 
        return f'({self._aabb.minX} {self._aabb.maxX} {self._aabb.minY} {self._aabb.maxY} {self._aabb.minZ} {self._aabb.maxZ})'
    
    def __repr__(self):
        cdef aabb3f _aabb = self.updateLocalAABB() 
        return f'AABB({self._aabb.minX} {self._aabb.maxX} {self._aabb.minY} {self._aabb.maxY} {self._aabb.minZ} {self._aabb.maxZ})'


cdef class AABBSet:
    """Represents a collection of axis-aligned bounding boxes (AABBs)
    
    :param count: number of bounding box
    :type count: int
    """
    cdef object __weakref__
    cdef GIM_AABB_SET _aabb_set
    
    def __cinit__(self, int count, flag=None):
        if flag is __create_uninitialized__:
            return
        
        gim_aabbset_alloc(&self._aabb_set, count)

    def __dealloc__(self):
        self._aabb_set.m_shared = 0
        gim_aabbset_destroy(&self._aabb_set)

    def __getitem__(self, index):
        if index < 0 or index >= self._aabb_set.m_count:
            raise IndexError('AABBSet index out of range')

        cdef aabb3f* temp = &self._aabb_set.m_boxes[index]
        aabb = AABB(temp[0].minX, temp[0].maxX, temp[0].minY, temp[0].maxY, temp[0].minZ, temp[0].maxZ)
        aabb._aabb_set = &self._aabb_set
        aabb._src_aabb = temp
        return aabb

    def __setitem__(self, index, bounds):
        if index < 0 or index >= self._aabb_set.m_count:
            raise IndexError('AABBSet index out of range')

        if isinstance(bounds, AABB) and bounds._id() == self._id():
            return
            
        self._aabb_set.m_boxes[index].minX = bounds[0]
        self._aabb_set.m_boxes[index].maxX = bounds[1]
        self._aabb_set.m_boxes[index].minY = bounds[2]
        self._aabb_set.m_boxes[index].maxY = bounds[3]
        self._aabb_set.m_boxes[index].minZ = bounds[4]
        self._aabb_set.m_boxes[index].maxZ = bounds[5]
        gim_aabbset_update(&self._aabb_set)

    @property
    def global_bounds(self):
        """Gets the AABB for the entire collection

        :return: bounding box of AABBSet
        :rtype: AABB
        """
        aabb = self._aabb_set.m_global_bound
        return AABB(aabb.minX, aabb.maxX, aabb.minY, aabb.maxY, aabb.minZ, aabb.maxZ)

    def find_intersections(self, aabb_set):
        """Finds all intersections between this AABBs of this 
        AABBSet and those of another.

        :param aabb_set: the other AABBSet
        :type aabb_set: AABBSet
        :return: A list of intersecting index pairs
        :rtype: List[Tuple[int]]
        """
        
        cdef GDYNAMIC_ARRAY gim_pairs
        GIM_CREATE_PAIR_SET(gim_pairs)

        cdef object id = aabb_set._id()
        
        gim_aabbset_bipartite_intersections(&self._aabb_set, <GIM_AABB_SET*>PyLong_AsVoidPtr(id), &gim_pairs)
        pairs = []
        
        cdef int count = gim_pairs.m_size
        cdef GIM_PAIR* data = <GIM_PAIR*>gim_pairs.m_pdata 
        
        for pair in data[:count]:
            pairs.append((pair.m_index1, pair.m_index2))
        
        GIM_DYNARRAY_DESTROY(gim_pairs)

        return pairs

    def __len__(self):
        return self._aabb_set.m_count
    
    def _id(self):
        return PyLong_FromVoidPtr(<void*>&self._aabb_set)


cdef class TriMesh:
    """Represents a triangle mesh
    
    :param vertices: array of 3D vertices
    :type vertices: Array[float]
    :param indices: array of indices
    :type indices: Array[int]
    """
    cdef object __weakref__
    cdef GIM_TRIMESH _trimesh
    cdef GBUFFER_MANAGER_DATA buffer_managers[G_BUFFER_MANAGER__MAX]
    cdef object _aabb_set
    
    def __cinit__(self, vertices, indices, flag=None):
        if flag is __create_uninitialized__:
            return
        
        if vertices is None or indices is None:
            raise ValueError('vertices and indices cannot be None.')     
        
        cdef float[:, ::1] _vertices = np.array(vertices, dtype=np.float32, copy=False, order='C',)
        cdef int[::1] _indices = np.array(indices, dtype=np.int32, copy=False, order='C',)
        
        if np.amin(indices) < 0 or np.amax(indices) >= vertices.shape[0]:
            raise ValueError("Vertex index out of range")
        
        gim_init_buffer_managers(self.buffer_managers)
        gim_trimesh_create_from_data(self.buffer_managers, &self._trimesh, <vec3f*>&_vertices[0, 0],
                                     _vertices.shape[0], 1, <GUINT32*>&_indices[0], _indices.size, 1, 1)
        self.initAABBSet()

    cdef initAABBSet(self):    
        gim_trimesh_update(&self._trimesh)

        self._trimesh.m_aabbset.m_shared = 1
        aabb_set = AABBSet(0, flag=__create_uninitialized__)
        aabb_set._aabb_set = self._trimesh.m_aabbset
        self._aabb_set = aabb_set
        
    def __dealloc__(self):
        gim_trimesh_destroy(&self._trimesh)
        gim_terminate_buffer_managers(self.buffer_managers)

    def transform(self, matrix not None):
        """Transform triangle mesh in-place with given matrix.
        
        :param matrix: 4 x 4 transformation matrix
        :type matrix: Array[float]
        """
        cdef float[:, ::1] _matrix = np.array(matrix, dtype=np.float32, copy=False, order='C',)
        if _matrix.shape[0] != 4 or _matrix.shape[1] != 4:
            raise ValueError("Transformation _matrix should have dimension (4, 4)")

        gim_trimesh_set_tranform(&self._trimesh, <mat4f>&_matrix[0, 0])
        gim_trimesh_update(&self._trimesh)
        cdef GIM_AABB_SET* aabb_set = <GIM_AABB_SET*>PyLong_AsVoidPtr(self._aabb_set._id())
        aabb_set[0].m_global_bound = self._trimesh.m_aabbset.m_global_bound

    @property
    def aabb_set(self):
        """Gets an AABBSet containing the AABBs of the triangles in trimesh

        :return: AABBSet of Trimesh
        :rtype: AABBSet
        """
        return self._aabb_set

    @property
    def bounds(self):
        """Gets an AABB bounding the trimesh

        :return: bounding box of Trimesh
        :rtype: AABB
        """
        return self._aabb_set.global_bounds

    def clone(self):
        """clones the trimesh.
        
        :return: trimesh clone
        :rtype: Trimesh
        """
        trimesh = TriMesh(None, None, flag=__create_uninitialized__)
        gim_init_buffer_managers(trimesh.buffer_managers)
        gim_trimesh_copy(&self._trimesh, trimesh.buffer_managers, &trimesh._trimesh, 0, 1)
        trimesh.initAABBSet()
        
        return trimesh
    
    @property
    def triangle_count(self):
        """Returns the number of triangles in the TriMesh.
        
        :return: the number of triangles in the TriMesh
        :rtype: int
        """
        return gim_trimesh_get_triangle_count(&self._trimesh)

    def triangle(self, int idx):
        """returns vertices of the triangle with specified index

        :param idx: triangle index
        :type idx: int
        :return: vertices of triangle with specified index
        :rtype: Tuple[Tuple[float, float, float]]
        """
        cdef vec3f v0 = [0., 0., 0.]
        cdef vec3f v1 = [0., 0., 0.]
        cdef vec3f v2 = [0., 0., 0.]
        
        gim_trimesh_locks_work_data(&self._trimesh)
        gim_trimesh_get_triangle_vertices(&self._trimesh, idx, v0, v1, v2)
        gim_trimesh_unlocks_work_data(&self._trimesh)

        return ((v0[0], v0[1], v0[2]),
                (v1[0], v1[1], v1[2]),
                (v2[0], v2[1], v2[2]))
    
    def decimate(self, int target_count):
        """Simplifies mesh to a given target number of faces. Returns the same mesh 
        if its face count is less than or equal to target_count.
        
        :param target_count: target number of faces
        :type target_count: int
        :return: simplified trimesh
        :rtype: Trimesh
        """
        if target_count >= self.triangle_count:
            return self

        cdef GBUFFER_ARRAY* vertex_buffer = &self._trimesh.m_source_vertex_buffer
        cdef GBUFFER_ARRAY* index_buffer = &self._trimesh.m_tri_index_buffer
        gim_buffer_array_lock(vertex_buffer, 1)
        gim_buffer_array_lock(index_buffer, 1)
        vertices = np.asarray(<float[:vertex_buffer.m_element_count, :3]><float*>vertex_buffer.m_buffer_data)
        indices = np.asarray(<int[:index_buffer.m_element_count]><int*>index_buffer.m_buffer_data)
        tmp = np.unique(vertices[indices], return_inverse=True, axis=0)
        cdef float[:, ::1] _vertices = tmp[0]
        cdef int[::1] _indices = tmp[1].astype(np.int32)  # np.unique returns int64
        read_mesh(&_vertices[0, 0], _vertices.shape[0], &_indices[0], indices.size)
        gim_buffer_array_unlock(vertex_buffer)
        gim_buffer_array_unlock(index_buffer)
        
        simplify_mesh(target_count, 7, False)
        
        cdef vector[float] v
        cdef vector[int] i
        write_mesh(v, i)

        return TriMesh(np.array(v, np.float32).reshape(-1, 3), np.array(i, np.int32))

    def _id(self):
        return PyLong_FromVoidPtr(<void*>&self._trimesh)


def trimesh_trimesh_collision(trimesh1, trimesh2, bool first_only=False):
    """Determines contacts of a trimesh-trimesh collision. 
    
    :param trimesh1: first triangle mesh
    :type trimesh1: Trimesh
    :param trimesh2: second triangle mesh
    :type trimesh2: Trimesh
    :param first_only: flag that indicates only first contact is required
    :type first_only: bool
    :return: a list of Contacts
    :rtype: List[Contact]
    """
    
    cdef GDYNAMIC_ARRAY gim_contacts
    GIM_CREATE_CONTACT_LIST(gim_contacts)

    cdef object id1 = trimesh1._id()
    cdef object id2 = trimesh2._id()
    cdef char mode = 1 if first_only else 0 
    gim_trimesh_trimesh_collision(<GIM_TRIMESH*>PyLong_AsVoidPtr(id1), <GIM_TRIMESH*>PyLong_AsVoidPtr(id2), &gim_contacts, mode)
       
    return extract_contact_data(gim_contacts)


def trimesh_sphere_collision(trimesh, center, float radius, bool first_only=False):
    """Determines contacts of a trimesh-sphere collision
    
    :param trimesh: triangle mesh
    :type trimesh: Trimesh
    :param center: center of the sphere
    :type center: Array[float]
    :param radius: radius of the sphere
    :type radius: float
    :param first_only: flag that indicates only first contact is required
    :type first_only: bool
    :return: a list of Contacts
    :rtype: List[Contact]
    """
    cdef GDYNAMIC_ARRAY gim_contacts
    GIM_CREATE_CONTACT_LIST(gim_contacts)

    cdef object id = trimesh._id()
    cdef vec3f _center = [center[0], center[1], center[2]]
    cdef char mode = 1 if first_only else 0 
    gim_trimesh_sphere_collision(<GIM_TRIMESH*>PyLong_AsVoidPtr(id), _center, radius, &gim_contacts, mode)
       
    return extract_contact_data(gim_contacts)


def trimesh_plane_collision(trimesh, plane, bool first_only=False):
    """Determines contacts of a trimesh-plane collision.
    
    :param trimesh: triangle mesh
    :type trimesh: Trimesh
    :param plane: plane parameters (a, b, c, d) in form of ax + by + cz + d = 0
    :type plane: Array[float]
    :param first_only: flag that indicates only first contact is required
    :type first_only: bool
    :return: a list of tuples containing point and penetration depth
    :rtype: List[Tuple[Array[float], float]]
    """
    
    cdef GDYNAMIC_ARRAY gim_contacts
    GIM_CREATE_TRIMESHPLANE_CONTACTS(gim_contacts)

    cdef object id = trimesh._id()
    cdef vec4f _plane = [plane[0], plane[1], plane[2], plane[3]]
    cdef char mode = 1 if first_only else 0 
    gim_trimesh_plane_collision(<GIM_TRIMESH*>PyLong_AsVoidPtr(id), _plane,  &gim_contacts, mode)
    contacts = []
    
    cdef int count = gim_contacts.m_size
    cdef vec4f* _contacts_data = <vec4f*>gim_contacts.m_pdata 
    
    for c in _contacts_data[:count]:
        t = np.array(<vec4f>c)
        contacts.append((t[0:3], t[3]))
    
    GIM_DYNARRAY_DESTROY(gim_contacts)

    return contacts


def trimesh_capsule_collision(trimesh, point1, point2, float radius, bool first_only=False):
    """Determines contacts of a trimesh-capsule collision.
    
    :param trimesh: triangle mesh
    :type trimesh: Trimesh
    :param point1: first end-point of the capsule
    :type point1: Array[float]
    :param point2: second end-point of the capsule
    :type point2: Array[float]
    :param radius: radius of the sphere
    :type radius: float
    :param first_only: flag that indicates only first contact is required
    :type first_only: bool
    :return: a list of Contacts
    :rtype: List[Contact]
    """
    cdef GDYNAMIC_ARRAY gim_contacts
    GIM_CREATE_CONTACT_LIST(gim_contacts)

    cdef object id = trimesh._id()
    cdef GIM_CAPSULE_DATA capsule
    capsule.m_radius = radius
    capsule.m_point1 = [point1[0], point1[1], point1[2]]
    capsule.m_point2 = [point2[0], point2[1], point2[2]]
    cdef char mode = 1 if first_only else 0 
    gim_trimesh_capsule_collision(<GIM_TRIMESH*>PyLong_AsVoidPtr(id), &capsule, &gim_contacts, mode)


    result = extract_contact_data(gim_contacts)

    return result[:1] if first_only else result


def trimesh_ray_collision(trimesh, origin, direction, float tmax):
    """Determines contact of a trimesh-ray collision. Collision is considered
    valid only when the ray collides with the front faces of the trimesh
    
    :param trimesh: triangle mesh
    :type trimesh: Trimesh
    :param origin: origin point of ray
    :type origin: Array[float]
    :param direction: direction vector of ray
    :type direction: Array[float]
    :param tmax: max distance param for ray.
    :type tmax: float
    :return: random contact if ray collides else None
    :rtype: Union[Contact, None]
    """

    cdef GIM_TRIANGLE_RAY_CONTACT_DATA gim_contact
    cdef vec3f _origin = [origin[0], origin[1], origin[2]]
    cdef vec3f _direction = [direction[0], direction[1], direction[2]]
    intersect = gim_trimesh_ray_collision(<GIM_TRIMESH*>PyLong_AsVoidPtr(trimesh._id()),
                                          _origin, _direction, tmax, &gim_contact)

    if intersect:
        pos= np.array(gim_contact.m_point)
        normal = np.array(gim_contact.m_normal)
        return Contact(pos, normal, gim_contact.tparam, gim_contact.m_face_id, None)


def trimesh_ray_closest_collision(trimesh, origin, direction, float tmax):
    """Determines closest contact of a trimesh-ray collision. Collision is considered
    valid only when the ray collides with the front faces of the trimesh
    
    :param trimesh: triangle mesh
    :type trimesh: Trimesh
    :param origin: origin point of ray
    :type origin: Array[float]
    :param direction: direction vector of ray
    :type direction: Array[float]
    :param tmax: max distance param for ray.
    :type tmax: float
    :return: closest contact if ray collides else None
    :rtype: Union[Contact, None]
    """
    
    cdef GIM_TRIANGLE_RAY_CONTACT_DATA gim_contact
    cdef vec3f _origin = [origin[0], origin[1], origin[2]]
    cdef vec3f _direction = [direction[0], direction[1], direction[2]]
    intersect = gim_trimesh_ray_closest_collision(<GIM_TRIMESH*>PyLong_AsVoidPtr(trimesh._id()),
                                                  _origin, _direction, tmax, &gim_contact)

    if intersect:
        pos= np.array(gim_contact.m_point)
        normal = np.array(gim_contact.m_normal)
        return Contact(pos, normal, gim_contact.tparam, gim_contact.m_face_id, None)


cdef extract_contact_data(GDYNAMIC_ARRAY  gim_contacts):
    contacts = []
    
    cdef int count = gim_contacts.m_size
    cdef GIM_CONTACT* _contacts_data = <GIM_CONTACT*>gim_contacts.m_pdata 
    
    for c in _contacts_data[:count]:
        pos= np.array(c.m_point)
        normal = np.array(c.m_normal)
        contacts.append(Contact(pos, normal, c.m_depth, c.m_feature1, c.m_feature2))
    
    GIM_DYNARRAY_DESTROY(gim_contacts) 

    return contacts


class Contact:
    """Represents a collision contact
    
    :param point: intersection point
    :type point: Array[float]
    :param normal: friction direction
    :type normal: Array[float]
    :param depth: peneration depth
    :type depth: float
    :param feature1: index of colliding triangle in first mesh.
    :type feature1: int
    :param feature2: index of colliding triangle in second mesh.
    :type feature2: int
    """
    def __init__(self, point, normal, depth, feature1, feature2):
        self.point = point
        self.normal = normal
        self.depth = depth
        self.feature1 = feature1
        self.feature2 = feature2
    
    def __str__(self):
        return (f'Contact: point: {self.point}, normal: {self.normal}, depth: {self.depth}, '
                f'feature1: {self.feature1}, feature2: {self.feature2}')

    def __repr__(self):
        return str(self)


cdef initialize():
    """init()"""

    gimpact_init()


initialize()



from libcpp cimport bool
from libcpp.vector cimport vector

cdef extern from "Python.h":
    object PyLong_FromVoidPtr(void *)
    void *PyLong_AsVoidPtr(object)


cdef extern from "GIMPACT/gimpact.h":
    void gimpact_init()
    void gimpact_terminate()


cdef extern from "GIMPACT/gim_math.h":
    ctypedef unsigned int GUINT32
    ctypedef int GINT32
    ctypedef float GREAL
    ctypedef GREAL vec3f[3]
    ctypedef GREAL vec4f[4]
    ctypedef GREAL mat4f[4][4]


cdef extern from "GIMPACT/gim_contact.h":
    cdef struct GIM_CONTACT:
        vec3f m_point
        vec3f m_normal
        GREAL m_depth
        void * m_handle1
        void * m_handle2
        GUINT32 m_feature1
        GUINT32 m_feature2

    cdef struct GIM_TRIANGLE_RAY_CONTACT_DATA:
        GREAL u
        GREAL v
        GREAL tparam
        GUINT32 m_face_id
        vec3f m_point
        vec3f m_normal

    void GIM_CREATE_CONTACT_LIST(GDYNAMIC_ARRAY contact_array)


cdef extern from "GIMPACT/gim_memory.h":
    cpdef enum:
        G_BUFFER_MANAGER_SYSTEM,
        G_BUFFER_MANAGER_SHARED,
        G_BUFFER_MANAGER__MAX   

    cdef struct GBUFFER_ID:
        pass
    
    cdef struct GBUFFER_ARRAY:
        GBUFFER_ID m_buffer_id
        char * m_buffer_data
        char m_byte_stride
        GUINT32 m_byte_offset
        GUINT32 m_element_count

    cdef struct GBUFFER_MANAGER_DATA:
        pass
    
    cdef struct GDYNAMIC_ARRAY:
        char * m_pdata
        GUINT32 m_size
        GUINT32 m_reserve_size
    
    void * gim_alloc(size_t size)

    GINT32 gim_buffer_array_lock(GBUFFER_ARRAY * array_data, int access)
    
    GINT32 gim_buffer_array_unlock(GBUFFER_ARRAY * array_data)

    void gim_init_buffer_managers(GBUFFER_MANAGER_DATA buffer_managers[])
    
    void gim_terminate_buffer_managers(GBUFFER_MANAGER_DATA buffer_managers[])

    void GIM_DYNARRAY_DESTROY(GDYNAMIC_ARRAY & array_data)


cdef extern from "GIMPACT/gim_boxpruning.h":
    cdef struct aabb3f:
        GREAL 	minX
        GREAL 	maxX
        GREAL 	minY
        GREAL 	maxY
        GREAL 	minZ
        GREAL 	maxZ

    struct GIM_RSORT_TOKEN:
        pass

    cdef struct GIM_AABB_SET:
        GUINT32 m_count
        aabb3f m_global_bound
        aabb3f * m_boxes
        GUINT32 * m_maxcoords
        GIM_RSORT_TOKEN * m_sorted_mincoords;
        char m_shared;
    
    cdef struct GIM_PAIR:
        GUINT32 m_index1
        GUINT32 m_index2

    void gim_aabbset_alloc(GIM_AABB_SET * aabbset, GUINT32 count)
    
    void gim_aabbset_destroy(GIM_AABB_SET * aabbset)

    void gim_aabbset_update(GIM_AABB_SET * aabbset) 	

    void gim_aabbset_bipartite_intersections(GIM_AABB_SET * aabbset1, GIM_AABB_SET * aabbset2, GDYNAMIC_ARRAY * collision_pairs)
    
    void GIM_CREATE_PAIR_SET(GDYNAMIC_ARRAY pair_array)

cdef extern from "GIMPACT/gim_boxpruning.h":
    void AABB_COPY(aabb3f& dest_aabb, aabb3f& src_aabb)
    
    void MERGEBOXES(aabb3f& destaabb, aabb3f& aabb)
    
    void BOXINTERSECTION(aabb3f& aabb1, aabb3f& aabb2, aabb3f& iaabb)
    
    void AABBCOLLISION(char& intersected, aabb3f& aabb1, aabb3f& aabb2) 


cdef extern from "GIMPACT/gim_tri_capsule_collision.h":
    cdef struct GIM_CAPSULE_DATA:
        GREAL m_radius
        vec3f m_point1
        vec3f m_point2


cdef extern from "GIMPACT/gim_trimesh.h": 
    cdef struct GIM_TRIMESH
    
    ctypedef void* (*gim_update_trimesh_function)(GIM_TRIMESH _trimesh)
   
    cdef struct GIM_TRIMESH:
        GBUFFER_ARRAY m_source_vertex_buffer
        GBUFFER_ARRAY m_tri_index_buffer
        char m_mask
        GBUFFER_ARRAY m_transformed_vertex_buffer
        GIM_AABB_SET m_aabbset
        GDYNAMIC_ARRAY m_planes_cache_buffer
        GDYNAMIC_ARRAY m_planes_cache_bitset
        gim_update_trimesh_function * m_update_callback
        mat4f m_transform

    void GIM_CREATE_TRIMESHPLANE_CONTACTS(GDYNAMIC_ARRAY contact_array)
    
    void gim_trimesh_create_from_data(GBUFFER_MANAGER_DATA buffer_managers[],
	            GIM_TRIMESH * trimesh, vec3f * vertex_array, GUINT32 vertex_count, char copy_vertices, 
	            GUINT32 * triindex_array, GUINT32 index_count,char copy_indices,char transformed_reply)
    
    void gim_trimesh_destroy(GIM_TRIMESH * trimesh)

    GUINT32 gim_trimesh_get_triangle_count(GIM_TRIMESH * trimesh)
    
    void gim_trimesh_set_tranform(GIM_TRIMESH * trimesh, mat4f transform)

    void gim_trimesh_get_triangle_vertices(GIM_TRIMESH * trimesh, GUINT32 triangle_index, vec3f v1,vec3f v2,vec3f v3)

    void gim_trimesh_locks_work_data(GIM_TRIMESH * trimesh)

    void gim_trimesh_unlocks_work_data(GIM_TRIMESH * trimesh)

    void gim_trimesh_copy(GIM_TRIMESH * source_trimesh, GBUFFER_MANAGER_DATA dest_buffer_managers[], GIM_TRIMESH * dest_trimesh, 
                          char copy_by_reference, char transformed_reply)	 		

    void gim_trimesh_update(GIM_TRIMESH * trimesh)

    void gim_trimesh_trimesh_collision(GIM_TRIMESH * trimesh1, GIM_TRIMESH * trimesh2, GDYNAMIC_ARRAY * contacts, char mode)
    
    void gim_trimesh_sphere_collision(GIM_TRIMESH * trimesh,vec3f center,GREAL radius, GDYNAMIC_ARRAY * contacts, char mode)
    
    void gim_trimesh_capsule_collision(GIM_TRIMESH * trimesh, GIM_CAPSULE_DATA * capsule, GDYNAMIC_ARRAY * contacts, char mode)

    void gim_trimesh_plane_collision(GIM_TRIMESH * trimesh, vec4f plane, GDYNAMIC_ARRAY * contacts, char mode)

    int gim_trimesh_ray_collision(GIM_TRIMESH * trimesh, vec3f origin, vec3f dir, GREAL tmax, GIM_TRIANGLE_RAY_CONTACT_DATA * contact)
    
    int gim_trimesh_ray_closest_collision(GIM_TRIMESH * trimesh, vec3f origin, vec3f dir, GREAL tmax, GIM_TRIANGLE_RAY_CONTACT_DATA * contact)


cdef extern from "simplify.h" namespace "Simplify":
    void read_mesh(float* vertices, int vertex_count, int* indices, int index_count)
    
    void simplify_mesh(int target_count, double agressiveness, bool verbose)
    
    void write_mesh(vector[float]& verts, vector[int]& indices)



import unittest
import numpy as np
import gimpact


def create_cubiod(width=10.0, height=10.0, depth=10.0):

    # centre cube at the origin
    w = width / 2
    h = height / 2
    d = depth / 2
    
    ftl = [-w, h, d] # front top left vertex
    ftr = [w, h, d] # front top right vertex
    fbl = [-w, -h, d] # front bottom left vertex
    fbr = [w, -h, d] # front bottom right vertex
    btl = [-w, h, -d] # back top left vertex
    btr = [w, h, -d] # back top right vertex
    bbl = [-w, -h, -d] # back bottom left vertex
    bbr = [w, -h, -d] # back bottom right vertex
    
    vertices = [fbl, fbr, ftl, ftr, ftl, ftr, 
                btl, btr, btl, btr, bbl, bbr, 
                bbl, bbr, fbl, fbr, fbr,  bbr, 
                ftr, btr, bbl, fbl, btl, ftl]

    indices =  np.array([0, 1, 2, 2, 1, 3, 
                4, 5, 6, 6, 5, 7, 
                8, 9, 10, 10, 9, 11, 
                12, 13, 14, 14, 13, 15, 
                16, 17, 18, 18, 17, 19, 
                20, 21, 22, 22, 21, 23])
    
    return {'vertices': np.array(vertices), 'indices': indices}


class TestAABB(unittest.TestCase):
    def testAABBExtents(self):
        aabb = gimpact.AABB(-1, 1, -1, 1, -1, 1)
        self.assertAlmostEqual(aabb.min_x, -1, 2)
        self.assertAlmostEqual(aabb.max_x, 1, 2)
        self.assertAlmostEqual(aabb.min_y, -1, 2)
        self.assertAlmostEqual(aabb.max_y, 1, 2)
        self.assertAlmostEqual(aabb.min_z, -1, 2)
        self.assertAlmostEqual(aabb.max_z, 1, 2)

        aabb.min_x = -0.5
        aabb.max_x = 0.5
        aabb.min_y = -0.5
        aabb.max_y = 0.5
        aabb.min_z = -0.5
        aabb.max_z = 0.5
        np.testing.assert_array_almost_equal(aabb.bounds, (-0.5, 0.5, -0.5, 0.5, -0.5, 0.5), decimal=2)
        aabb.bounds = (1, -2, 3, -4, -5, 6)
        np.testing.assert_array_almost_equal(aabb.bounds, (1, -2, 3, -4, -5, 6), decimal=2)
        self.assertAlmostEqual(aabb.min_x, 1, 2)
        self.assertAlmostEqual(aabb.max_x, -2, 2)
        self.assertAlmostEqual(aabb.min_y, 3, 2)
        self.assertAlmostEqual(aabb.max_y, -4, 2)
        self.assertAlmostEqual(aabb.min_z, -5, 2)
        self.assertAlmostEqual(aabb.max_z, 6, 2)

    def testAABBMerge(self):
        aabb1 = gimpact.AABB(-1, 1, -1, 1, -1, 1)
        aabb2 = gimpact.AABB(-1, 1, -1, 1, 1.5, 2)
        aabb1.merge(aabb2)
        np.testing.assert_array_almost_equal(aabb1.bounds, (-1, 1, -1, 1, -1, 2), decimal=2)

    def testAABBIntersection(self):
        aabb1 = gimpact.AABB(-1, 1, -1, 1, -1, 1)
        aabb2 = gimpact.AABB(-1, 1, -1, 1, 1.5, 2)
        self.assertFalse(aabb1.intersects(aabb2))
        
        aabb1 = gimpact.AABB(-1, 1, -1, 0.5, -1, 2.5)
        aabb2 = gimpact.AABB(-1, 1, -1, 1, -1.5, 2)
        self.assertTrue(aabb1.intersects(aabb2))
        np.testing.assert_array_almost_equal(aabb1.intersection(aabb2).bounds, (-1, 1, -1, 0.5, -1, 2), decimal=2)
    
    def testClone(self):
        aabb1 = gimpact.AABB(-1, 1, -1, 1, -1, 1)
        aabb2 = aabb1.clone()
        np.testing.assert_array_almost_equal(aabb1.bounds, aabb2.bounds, decimal=2)
        aabb2.max_z = 5
        self.assertNotAlmostEqual(aabb1.max_z, aabb2.max_z)


class TestAABBSet(unittest.TestCase):
    def testBounds(self):
        aabb_set = gimpact.AABBSet(10)
        self.assertEqual(len(aabb_set), 10)

        for aabb in aabb_set:
            aabb.bounds = (-9., 0., 0., 0., 0., 10.)
        aabb_set[0] = (0., 6., 0., 0., -5., 0.)
        np.testing.assert_array_almost_equal(aabb_set.global_bounds.bounds, (-9., 6., 0., 0., -5., 10.), decimal=2)

        bounds = aabb.bounds
        del aabb_set
        np.testing.assert_array_almost_equal(aabb.bounds, bounds, decimal=2)

    def testIntersection(self):
        aabb_set = gimpact.AABBSet(2)
        aabb_set[0] = (-1, 1, -1, 1, -1, 1)
        aabb_set[1] = gimpact.AABB(-1, 1, -1, 1, 1.5, 2)

        np.testing.assert_array_almost_equal(aabb_set.global_bounds.bounds, (-1.,  1., -1.,  1., -1.,  2.), decimal=2)

        expected = [(0, 0), (1, 1)]
        pairs = aabb_set.find_intersections(aabb_set)
        self.assertEqual(len(pairs), 2)
        for pair in pairs:
            self.assertIn(pair, expected)
        
        aabb_set2 = gimpact.AABBSet(2)
        aabb_set2[0] = gimpact.AABB(-1, 1, -1, 1, 1.5, 2)
        aabb_set2[1] = (-1, 1, -1, 1, -1, 1)
        expected = [(0, 1), (1, 0)]
        pairs = aabb_set2.find_intersections(aabb_set)
        self.assertEqual(len(pairs), 2)
        for pair in pairs:
            self.assertIn(pair, expected)


class TestTrimesh(unittest.TestCase):
    def setUp(self):
        mesh = create_cubiod()
        self.vertices = mesh['vertices']
        self.indices = mesh['indices']
    
    def testDecimation(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)
        trimesh2 = trimesh.decimate(8)
        self.assertEqual(trimesh2.triangle_count, 8)

    def testTriangle(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)
        self.assertEqual(trimesh.triangle_count, self.indices.size//3)
        
        for i in range(trimesh.triangle_count):
            np.testing.assert_array_almost_equal(trimesh.triangle(i), 
                                                 self.vertices[self.indices[i*3:i*3+3]], decimal=2)

        trimesh2 = trimesh.clone()
        for i in range(trimesh.triangle_count):
            a = np.array(trimesh.triangle(i))
            b = np.array(trimesh2.triangle(i))
            np.testing.assert_array_almost_equal(a, b, decimal=2)

    def testTransform(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)
        t = np.identity(4)
        t[0:3, 3] = [10, 10, 10]
        trimesh.transform(t)
        v = self.vertices[self.indices]
        v = v @ t[0:3, 0:3].transpose() + t[0:3, 3] 
        for i in range(trimesh.triangle_count):
            np.testing.assert_array_almost_equal(trimesh.triangle(i), v[i*3:i*3+3], decimal=2)

        trimesh2 = trimesh.clone()
        t = np.identity(4)
        t[0:2, 0:2] = [[0, -1], [1, 0]]
        trimesh2.transform(t)
        v = self.vertices[self.indices]
        v = v @ t[0:3, 0:3].transpose() 
        for i in range(trimesh2.triangle_count):
            np.testing.assert_array_almost_equal(trimesh2.triangle(i), v[i*3:i*3+3], decimal=2)

    def testAABBSet(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)
        aabb_set = trimesh.aabb_set
        np.testing.assert_array_almost_equal(aabb_set.global_bounds.bounds, trimesh.bounds.bounds, decimal=2)

        t = trimesh.clone()
        aabb_set = t.aabb_set
        self.assertEqual(trimesh.triangle_count, t.triangle_count)
        del t
        self.assertEqual(len(trimesh.aabb_set), len(aabb_set))
    
    def testTrimeshCollision(self):
        trimesh1 = gimpact.TriMesh(self.vertices, self.indices)
        trimesh2 = trimesh1.clone()
    
        contacts = gimpact.trimesh_trimesh_collision(trimesh1, trimesh2, True)
        self.assertEqual(len(contacts), 1)

        t = np.identity(4)
        t[2, 3] = 11
        trimesh2.transform(t)
        contacts = gimpact.trimesh_trimesh_collision(trimesh1, trimesh2)
        self.assertEqual(len(contacts), 0)

        t[:3, 3] = [10, 10, 10]
        trimesh2.transform(t)
        contacts = gimpact.trimesh_trimesh_collision(trimesh1, trimesh2)
        np.testing.assert_array_almost_equal(contacts[0].point, [5, 5, 5], decimal=2)
        np.testing.assert_array_almost_equal(contacts[0].normal, [0, 0, -1], decimal=2)
        self.assertAlmostEqual(contacts[0].depth, 0, 2)
        self.assertEqual(contacts[0].feature1, 1)
        self.assertEqual(contacts[0].feature2, 4)

    def testCapsuleCollision(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)
        
        radius = 1
        start = [-1., 0., 5.]
        stop = [1., 0., 5.]
        contacts = gimpact.trimesh_capsule_collision(trimesh, start, stop, radius, True)
        self.assertEqual(len(contacts), 1)
        
        contacts = gimpact.trimesh_capsule_collision(trimesh, start, stop, radius)
        self.assertEqual(len(contacts), 3)
        np.testing.assert_array_almost_equal(contacts[0].point, [-1, 0, 4], decimal=2)
        np.testing.assert_array_almost_equal(contacts[0].normal, [0, 0, -1], decimal=2)
        self.assertAlmostEqual(contacts[0].depth, 1, 2)
        self.assertEqual(contacts[0].feature1, 0)

        np.testing.assert_array_almost_equal(contacts[1].point, [0, 0, 4], decimal=2)
        np.testing.assert_array_almost_equal(contacts[1].normal, [0, 0, -1], decimal=2)
        self.assertAlmostEqual(contacts[1].depth, 1, 2)
        self.assertEqual(contacts[1].feature1, 1)
 
        np.testing.assert_array_almost_equal(contacts[2].point, [1, 0, 4], decimal=2)
        np.testing.assert_array_almost_equal(contacts[2].normal, [0, 0, -1], decimal=2)
        self.assertAlmostEqual(contacts[2].depth, 1, 2)
        self.assertEqual(contacts[2].feature1, 1)

    def testPlaneCollision(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)

        contacts = gimpact.trimesh_plane_collision(trimesh, np.array([0., 0., 1., 0.]), True)
        self.assertEqual(len(contacts), 1)
        contacts = gimpact.trimesh_plane_collision(trimesh, np.array([0., 0., 1., 0.]))
        self.assertEqual(len(contacts), self.vertices.shape[0]/2)
        for contact, depth in contacts:
            self.assertAlmostEqual(contact[2], -5, 2)
            self.assertAlmostEqual(depth, 5, 2)

        contacts = gimpact.trimesh_plane_collision(trimesh, np.array([0., 0., 1., -5.01]))
        self.assertEqual(len(contacts), 0)
        contacts = gimpact.trimesh_plane_collision(trimesh, np.array([0., 0., -1., 5.01]))
        self.assertEqual(len(contacts), self.vertices.shape[0])     

    def testRayCollision(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)
        start = [0., 0., 0.]
        direction = [-1., 0., 0.]
        depth = 100

        contact = gimpact.trimesh_ray_collision(trimesh, start, direction, depth)
        # No contact because ray originates from inside the object
        self.assertIsNone(contact)
        
        contact = gimpact.trimesh_ray_collision(trimesh, [0., 5.1, 0.], direction, depth)
        self.assertIsNone(contact)
        
        contact = gimpact.trimesh_ray_collision(trimesh, [2.5, -6.0, 0.], [0., 1., 0.], depth)
        np.testing.assert_array_almost_equal(contact.point, [2.5, -5, 0], decimal=2)
        np.testing.assert_array_almost_equal(contact.normal, [0, 1, 0], decimal=2)
        self.assertAlmostEqual(contact.depth, 1, 2)
        self.assertEqual(contact.feature1, 7)

        contact = gimpact.trimesh_ray_closest_collision(trimesh, np.array([5., 0., 5.]), np.array([0., 0., -1.]), 100)
        np.testing.assert_array_almost_equal(contact.point, [5, 0, 5], decimal=2)
        np.testing.assert_array_almost_equal(contact.normal, [0, 0, -1], decimal=2)
        self.assertAlmostEqual(contact.depth, 0, 2)
        self.assertEqual(contact.feature1, 1)

    def testSphereCollision(self):
        trimesh = gimpact.TriMesh(self.vertices, self.indices)

        radius = 1.0   
        center = [5., 5., 5.]    
        contacts = gimpact.trimesh_sphere_collision(trimesh, center, radius, True)
        self.assertEqual(len(contacts), 1)
        
        contacts = gimpact.trimesh_sphere_collision(trimesh, center, radius)
        self.assertEqual(len(contacts), 3)
        np.testing.assert_array_almost_equal(contacts[0].point, [5, 5, 4], decimal=2)
        np.testing.assert_array_almost_equal(contacts[0].normal, [0, 0, -1], decimal=2)
        self.assertAlmostEqual(contacts[0].depth, 1, 2)
        self.assertEqual(contacts[0].feature1, 1)

        np.testing.assert_array_almost_equal(contacts[1].point, [5, 4, 5], decimal=2)
        np.testing.assert_array_almost_equal(contacts[1].normal, [0, -1, 0], decimal=2)
        self.assertAlmostEqual(contacts[1].depth, 1, 2)
        self.assertEqual(contacts[1].feature1, 3)
 
        np.testing.assert_array_almost_equal(contacts[2].point, [4, 5, 5], decimal=2)
        np.testing.assert_array_almost_equal(contacts[2].normal, [-1, 0, 0], decimal=2)
        self.assertAlmostEqual(contacts[2].depth, 1, 2)
        self.assertEqual(contacts[2].feature1, 8)


if __name__ == '__main__':
    unittest.main()



import os
from setuptools import setup
from setuptools.command.build_ext import build_ext
from setuptools.extension import Extension

# Package meta-data.
NAME = 'gimpact'
DESCRIPTION = 'An unofficial python extension for the GImpact collision library.'
URL = 'https://github.com/StephenNneji'
AUTHOR = 'Stephen Nneji'
EMAIL = 'steve.nneji@gmail.com'
VERSION = '0.1.1'

path=os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(path, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


class build_ext(build_ext):
    def build_extensions(self):
        import numpy
        try:
            from Cython.Build import cythonize
            BUILD_CYTHON = True
        except ImportError:
            BUILD_CYTHON = False
            
        source = 'gimpact.pyx' if BUILD_CYTHON else 'gimpact.cpp'

        modules = Extension(
            name='gimpact',
            sources=[source, 'GIMPACT/src/gimpact.cpp', 'GIMPACT/src/gim_contact.cpp', 'GIMPACT/src/gim_math.cpp', 
                    'GIMPACT/src/gim_boxpruning.cpp', 'GIMPACT/src/gim_memory.cpp','GIMPACT/src/gim_tri_tri_overlap.cpp',
                    'GIMPACT/src/gim_trimesh.cpp','GIMPACT/src/gim_trimesh_capsule_collision.cpp', 'GIMPACT/src/gim_trimesh_ray_collision.cpp',
                    'GIMPACT/src/gim_trimesh_sphere_collision.cpp', 'GIMPACT/src/gim_trimesh_trimesh_collision.cpp'],
            language='c++',
            include_dirs=['GIMPACT/include', numpy.get_include()],
        )

        self.distribution.ext_modules[:] = cythonize([modules]) if BUILD_CYTHON else [modules]

        self.swig_opts = None
        super().finalize_options()

        super().build_extensions()


setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type="text/markdown",
      url=URL,
      author=AUTHOR,
      author_email=EMAIL,
      license='BSD',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Topic :: Games/Entertainment :: Simulation',
          'Topic :: Scientific/Engineering :: Physics',
          'License :: OSI Approved :: BSD License',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Cython'
      ],
      keywords='GImpact,Trimesh,Collision detection,Cython',
      cmdclass={"build_ext": build_ext},
      ext_modules=[Extension("", [])],
      test_suite='tests',
      install_requires = ['numpy>=1.16.4'],
      tests_require=['numpy>=1.16.4'],
)
